DEFAULT_SEARCH_OPTIONS = {
    "limit": 100,
    "offset": 0,
    "sortBy": {
        "column": "name",
        "order": "asc",
    },
}
DEFAULT_FILE_OPTIONS = {
    "cache-control": "3600",
    "content-type": "text/plain;charset=UTF-8",
    "x-upsert": "false",
}

DEFAULT_TIMEOUT = 20
